@@ -31,6 +31,7 @@ list active_proto_list;
 static list inactive_proto_list;
 static list initial_proto_list;
 static list flush_proto_list;
+static struct proto *initial_device_proto;
 
 static event *proto_flush_event;
 
@@ -248,6 +249,8 @@ protos_postconfig(struct config *c)
   DBG("\n");
 }
 
+extern struct protocol proto_unix_iface;
+
 static struct proto *
 proto_init(struct proto_config *c)
 {
@@ -257,6 +260,9 @@ proto_init(struct proto_config *c)
   q->proto_state = PS_DOWN;
   q->core_state = FS_HUNGRY;
   proto_enqueue(&initial_proto_list, q);
+  if (p == &proto_unix_iface)
+    initial_device_proto = q;
+
   add_tail(&proto_list, &q->glob_node);
   PD(q, "Initializing%s", q->disabled ? " [disabled]" : "");
   return q;
@@ -271,7 +277,7 @@ proto_reconfigure(struct proto *p, struct proto_config *oc, struct proto_config
 
   /* If there is a too big change in core attributes, ... */
   if ((nc->protocol != oc->protocol) ||
-      (nc->disabled != oc->disabled) ||
+      (nc->disabled != p->disabled) ||
       (nc->table->table != oc->table->table) ||
       (proto_get_router_id(nc) != proto_get_router_id(oc)))
     return 0;
@@ -413,6 +419,14 @@ protos_commit(struct config *new, struct config *old, int force_reconfig, int ty
   DBG("\tdone\n");
 
   DBG("Protocol start\n");
+
+  /* Start device protocol first */
+  if (initial_device_proto)
+  {
+    proto_rethink_goal(initial_device_proto);
+    initial_device_proto = NULL;
+  }
+
   WALK_LIST_DELSAFE(p, n, initial_proto_list)
     proto_rethink_goal(p);
 }
@@ -546,6 +560,9 @@ protos_build(void)
   init_list(&initial_proto_list);
   init_list(&flush_proto_list);
   proto_build(&proto_device);
+#ifdef CONFIG_RADV
+  proto_build(&proto_radv);
+#endif
 #ifdef CONFIG_RIP
   proto_build(&proto_rip);
 #endif
@@ -577,10 +594,8 @@ proto_fell_down(struct proto *p)
   bzero(&p->stats, sizeof(struct proto_stats));
   rt_unlock_table(p->table);
 
-#ifdef CONFIG_PIPE
-  if (proto_is_pipe(p))
-    rt_unlock_table(pipe_get_peer_table(p));
-#endif
+  if (p->proto->cleanup)
+    p->proto->cleanup(p);
 
   proto_rethink_goal(p);
 }
@@ -740,8 +755,6 @@ proto_notify_state(struct proto *p, unsigned ps)
     }
 }
 
-extern struct protocol proto_unix_iface;
-
 static void
 proto_flush_all(void *unused UNUSED)
 {
